# Copyright (c) 2014 Mirantis, Inc.
# Copyright 2020 Nokia Software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from tempest.lib import exceptions

from mistralclient.tests.functional.cli import base
from mistralclient.tests.functional.cli.v2 import base_v2


MISTRAL_URL = "http://localhost:8989/v2"


class SimpleMistralCLITests(base.MistralCLIAuth):
    """Basic tests, check '-list', '-help' commands."""

    _mistral_url = MISTRAL_URL

    @classmethod
    def setUpClass(cls):
        super(SimpleMistralCLITests, cls).setUpClass()

    def test_workbooks_list(self):
        workbooks = self.parser.listing(self.mistral('workbook-list'))

        self.assertTableStruct(
            workbooks,
            ['Name', 'Tags', 'Created at', 'Updated at']
        )

    def test_workflow_list(self):
        workflows = self.parser.listing(self.mistral('workflow-list'))

        self.assertTableStruct(
            workflows,
            ['ID', 'Name', 'Tags', 'Input', 'Scope', 'Created at',
             'Updated at']
        )

    def test_executions_list(self):
        executions = self.parser.listing(self.mistral('execution-list'))

        self.assertTableStruct(
            executions,
            ['ID', 'Workflow name', 'Workflow ID', 'State', 'Created at',
             'Updated at']
        )

    def test_tasks_list(self):
        tasks = self.parser.listing(self.mistral('task-list'))

        self.assertTableStruct(
            tasks,
            ['ID', 'Name', 'Workflow name', 'Workflow Execution ID', 'State']
        )

    def test_cron_trigger_list(self):
        triggers = self.parser.listing(self.mistral('cron-trigger-list'))

        self.assertTableStruct(
            triggers,
            ['Name', 'Workflow', 'Pattern', 'Next execution time',
             'Remaining executions', 'Created at', 'Updated at']
        )

    def test_event_trigger_list(self):
        triggers = self.parser.listing(self.mistral('event-trigger-list'))

        self.assertTableStruct(
            triggers,
            ['ID', 'Name', 'Workflow ID', 'Exchange', 'Topic', 'Event',
             'Created at', 'Updated at']
        )

    def test_actions_list(self):
        actions = self.parser.listing(self.mistral('action-list'))

        self.assertTableStruct(
            actions,
            ['Name', 'Is system', 'Input', 'Description',
             'Tags', 'Created at', 'Updated at']
        )

    def test_environments_list(self):
        envs = self.parser.listing(self.mistral('environment-list'))

        self.assertTableStruct(
            envs,
            ['Name', 'Description', 'Scope', 'Created at', 'Updated at']
        )

    def test_action_execution_list(self):
        act_execs = self.parser.listing(self.mistral('action-execution-list'))

        self.assertTableStruct(
            act_execs,
            ['ID', 'Name', 'Workflow name', 'State', 'Accepted']
        )

    def test_action_execution_list_with_limit(self):
        act_execs = self.parser.listing(
            self.mistral(
                'action-execution-list',
                params='--limit 1'
            )
        )

        self.assertEqual(1, len(act_execs))


class WorkbookCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with workbooks."""

    @classmethod
    def setUpClass(cls):
        super(WorkbookCLITests, cls).setUpClass()

    def test_workbook_create_delete(self):
        wb = self.mistral_admin('workbook-create', params=self.wb_def)

        wb_name = self.get_field_value(wb, "Name")

        self.assertTableStruct(wb, ['Field', 'Value'])

        wbs = self.mistral_admin('workbook-list')
        self.assertIn(wb_name, [w['Name'] for w in wbs])

        wbs = self.mistral_admin('workbook-list')
        self.assertIn(wb_name, [w['Name'] for w in wbs])

        self.mistral_admin('workbook-delete', params=wb_name)

        wbs = self.mistral_admin('workbook-list')
        self.assertNotIn(wb_name, [w['Name'] for w in wbs])

    def test_workbook_create_with_tags(self):
        wb = self.workbook_create(self.wb_with_tags_def)

        self.assertIn('tag', self.get_field_value(wb, 'Tags'))

    def test_workbook_update(self):
        wb = self.workbook_create(self.wb_def)

        wb_name = self.get_field_value(wb, "Name")

        init_update_at = self.get_field_value(wb, "Updated at")
        tags = self.get_field_value(wb, 'Tags')

        self.assertNotIn('tag', tags)

        wb = self.mistral_admin('workbook-update', params=self.wb_def)

        update_at = self.get_field_value(wb, "Updated at")
        name = self.get_field_value(wb, 'Name')
        tags = self.get_field_value(wb, 'Tags')

        self.assertEqual(wb_name, name)
        self.assertNotIn('tag', tags)
        self.assertEqual(init_update_at, update_at)

        wb = self.mistral_admin(
            'workbook-update',
            params=self.wb_with_tags_def
        )

        self.assertTableStruct(wb, ['Field', 'Value'])

        update_at = self.get_field_value(wb, "Updated at")
        name = self.get_field_value(wb, 'Name')
        tags = self.get_field_value(wb, 'Tags')

        self.assertEqual(wb_name, name)
        self.assertIn('tag', tags)
        self.assertNotEqual(init_update_at, update_at)

    def test_workbook_get(self):
        created = self.workbook_create(self.wb_with_tags_def)

        wb_name = self.get_field_value(created, "Name")

        fetched = self.mistral_admin('workbook-get', params=wb_name)

        created_wb_name = self.get_field_value(created, 'Name')
        fetched_wb_name = self.get_field_value(fetched, 'Name')

        self.assertEqual(created_wb_name, fetched_wb_name)

        created_wb_tag = self.get_field_value(created, 'Tags')
        fetched_wb_tag = self.get_field_value(fetched, 'Tags')

        self.assertEqual(created_wb_tag, fetched_wb_tag)

    def test_workbook_get_definition(self):
        wb = self.workbook_create(self.wb_def)

        wb_name = self.get_field_value(wb, "Name")

        definition = self.mistral_admin(
            'workbook-get-definition',
            params=wb_name
        )

        self.assertNotIn('404 Not Found', definition)

    def test_workbook_validate_with_valid_def(self):
        wb = self.mistral_admin('workbook-validate', params=self.wb_def)

        wb_valid = self.get_field_value(wb, 'Valid')
        wb_error = self.get_field_value(wb, 'Error')

        self.assertEqual('True', wb_valid)
        self.assertEqual('None', wb_error)

    def test_workbook_validate_with_invalid_def(self):
        self.create_file('wb.yaml', 'name: wb\n')

        wb = self.mistral_admin('workbook-validate', params='wb.yaml')

        wb_valid = self.get_field_value(wb, 'Valid')
        wb_error = self.get_field_value(wb, 'Error')

        self.assertEqual('False', wb_valid)
        self.assertNotEqual('None', wb_error)


class WorkflowCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with workflows."""

    @classmethod
    def setUpClass(cls):
        super(WorkflowCLITests, cls).setUpClass()

    def test_workflow_create_delete(self):
        init_wfs = self.mistral_admin('workflow-create', params=self.wf_def)

        wf_names = [wf['Name'] for wf in init_wfs]

        self.assertTableStruct(init_wfs, ['Name', 'Created at', 'Updated at'])

        wfs = self.mistral_admin('workflow-list')
        self.assertIn(wf_names[0], [workflow['Name'] for workflow in wfs])

        for wf_name in wf_names:
            self.mistral_admin('workflow-delete', params=wf_name)

        wfs = self.mistral_admin('workflow-list')
        for wf in wf_names:
            self.assertNotIn(wf, [workflow['Name'] for workflow in wfs])

    def test_workflow_within_namespace_create_delete(self):
        params = self.wf_def + ' --namespace abcdef'
        init_wfs = self.mistral_admin('workflow-create', params=params)

        wf_names = [wf['Name'] for wf in init_wfs]

        self.assertTableStruct(init_wfs, ['Name', 'Created at', 'Updated at'])

        wfs = self.mistral_admin('workflow-list')
        self.assertIn(wf_names[0], [workflow['Name'] for workflow in wfs])

        for wf_name in wf_names:
            self.mistral_admin(
                'workflow-delete',
                params=wf_name + ' --namespace abcdef'
            )

        wfs = self.mistral_admin('workflow-list')
        for wf in wf_names:
            self.assertNotIn(wf, [workflow['Name'] for workflow in wfs])

        init_wfs = self.mistral_admin('workflow-create', params=params)
        wf_ids = [wf['ID'] for wf in init_wfs]
        for wf_id in wf_ids:
            self.mistral_admin('workflow-delete', params=wf_id)

        for wf in wf_names:
            self.assertNotIn(wf, [workflow['Name'] for workflow in wfs])

    def test_create_wf_with_tags(self):
        init_wfs = self.workflow_create(self.wf_def)
        wf_name = init_wfs[1]['Name']

        self.assertTableStruct(
            init_wfs,
            ['Name', 'Created at', 'Updated at', 'Tags']
        )

        created_wf_info = self.get_item_info(
            get_from=init_wfs,
            get_by='Name',
            value=wf_name
        )

        self.assertEqual('tag', created_wf_info['Tags'])

    def test_workflow_update(self):
        wf = self.workflow_create(self.wf_def)

        wf_name = wf[0]['Name']
        wf_id = wf[0]['ID']

        created_wf_info = self.get_item_info(
            get_from=wf,
            get_by='Name',
            value=wf_name
        )

        # Update a workflow with definition unchanged.
        upd_wf = self.mistral_admin(
            'workflow-update',
            params='{0}'.format(self.wf_def)
        )

        self.assertTableStruct(upd_wf, ['Name', 'Created at', 'Updated at'])

        updated_wf_info = self.get_item_info(
            get_from=upd_wf,
            get_by='Name',
            value=wf_name
        )

        self.assertEqual(wf_name, upd_wf[0]['Name'])
        self.assertEqual(
            created_wf_info['Created at'].split(".")[0],
            updated_wf_info['Created at']
        )
        self.assertEqual(
            created_wf_info['Updated at'],
            updated_wf_info['Updated at']
        )

        # Update a workflow with definition changed.
        upd_wf = self.mistral_admin(
            'workflow-update',
            params='{0}'.format(self.wf_with_delay_def)
        )

        self.assertTableStruct(upd_wf, ['Name', 'Created at', 'Updated at'])

        updated_wf_info = self.get_item_info(
            get_from=upd_wf,
            get_by='Name',
            value=wf_name
        )

        self.assertEqual(wf_name, upd_wf[0]['Name'])
        self.assertEqual(
            created_wf_info['Created at'].split(".")[0],
            updated_wf_info['Created at']
        )
        self.assertNotEqual(
            created_wf_info['Updated at'],
            updated_wf_info['Updated at']
        )

        # Update a workflow with uuid.
        upd_wf = self.mistral_admin(
            'workflow-update',
            params='{0} --id {1}'.format(self.wf_with_delay_def, wf_id)
        )

        self.assertTableStruct(upd_wf, ['Name', 'Created at', 'Updated at'])

        updated_wf_info = self.get_item_info(
            get_from=upd_wf,
            get_by='ID',
            value=wf_id
        )

        self.assertEqual(wf_name, upd_wf[0]['Name'])
        self.assertEqual(
            created_wf_info['Created at'].split(".")[0],
            updated_wf_info['Created at']
        )
        self.assertNotEqual(
            created_wf_info['Updated at'],
            updated_wf_info['Updated at']
        )

    def test_workflow_update_within_namespace(self):
        namespace = 'abc'
        wf = self.workflow_create(self.wf_def, namespace=namespace)

        wf_name = wf[0]['Name']
        wf_id = wf[0]['ID']
        wf_namespace = wf[0]['Namespace']
        created_wf_info = self.get_item_info(
            get_from=wf,
            get_by='Name',
            value=wf_name
        )

        # Update a workflow with definition unchanged.
        upd_wf = self.mistral_admin(
            'workflow-update',
            params='{0} --namespace {1}'.format(self.wf_def, namespace)
        )

        self.assertTableStruct(upd_wf, ['Name', 'Created at', 'Updated at'])

        updated_wf_info = self.get_item_info(
            get_from=upd_wf,
            get_by='Name',
            value=wf_name
        )

        self.assertEqual(wf_name, upd_wf[0]['Name'])
        self.assertEqual(namespace, wf_namespace)
        self.assertEqual(wf_namespace, upd_wf[0]['Namespace'])
        self.assertEqual(
            created_wf_info['Created at'].split(".")[0],
            updated_wf_info['Created at']
        )
        self.assertEqual(
            created_wf_info['Updated at'],
            updated_wf_info['Updated at']
        )

        # Update a workflow with definition changed.
        upd_wf = self.mistral_admin(
            'workflow-update',
            params='{0} --namespace {1}'.format(
                self.wf_with_delay_def,
                namespace
            )
        )

        self.assertTableStruct(upd_wf, ['Name', 'Created at', 'Updated at'])

        updated_wf_info = self.get_item_info(
            get_from=upd_wf,
            get_by='Name',
            value=wf_name
        )

        self.assertEqual(wf_name, upd_wf[0]['Name'])
        self.assertEqual(
            created_wf_info['Created at'].split(".")[0],
            updated_wf_info['Created at']
        )
        self.assertNotEqual(
            created_wf_info['Updated at'],
            updated_wf_info['Updated at']
        )

        # Update a workflow with uuid.
        upd_wf = self.mistral_admin(
            'workflow-update',
            params='{0} --id {1}'.format(self.wf_with_delay_def, wf_id)
        )

        self.assertTableStruct(upd_wf, ['Name', 'Created at', 'Updated at'])

        updated_wf_info = self.get_item_info(
            get_from=upd_wf,
            get_by='ID',
            value=wf_id
        )

        self.assertEqual(wf_name, upd_wf[0]['Name'])
        self.assertEqual(
            created_wf_info['Created at'].split(".")[0],
            updated_wf_info['Created at']
        )
        self.assertNotEqual(
            created_wf_info['Updated at'],
            updated_wf_info['Updated at']
        )

    def test_workflow_update_truncate_input(self):
        input_value = "very_long_input_parameter_name_that_should_be_truncated"

        wf_def = """
        version: "2.0"
        workflow1:
          input:
            - {0}
          tasks:
            task1:
              action: std.noop
        """.format(input_value)

        self.create_file('wf.yaml', wf_def)
        self.workflow_create('wf.yaml')

        updated_wf = self.mistral_admin('workflow-update', params='wf.yaml')

        updated_wf_info = self.get_item_info(
            get_from=updated_wf,
            get_by='Name',
            value='workflow1'
        )

        self.assertEqual(updated_wf_info['Input'][:-3], input_value[:25])

    def test_workflow_get(self):
        created = self.workflow_create(self.wf_def)

        wf_name = created[0]['Name']

        fetched = self.mistral_admin('workflow-get', params=wf_name)
        fetched_wf_name = self.get_field_value(fetched, 'Name')

        self.assertEqual(wf_name, fetched_wf_name)

    def test_workflow_get_with_id(self):
        created = self.workflow_create(self.wf_def)

        wf_name = created[0]['Name']
        wf_id = created[0]['ID']

        fetched = self.mistral_admin('workflow-get', params=wf_id)
        fetched_wf_name = self.get_field_value(fetched, 'Name')

        self.assertEqual(wf_name, fetched_wf_name)

    def test_workflow_get_definition(self):
        wf = self.workflow_create(self.wf_def)

        wf_name = wf[0]['Name']

        definition = self.mistral_admin(
            'workflow-get-definition',
            params=wf_name
        )

        self.assertNotIn('404 Not Found', definition)

    def test_workflow_validate_with_valid_def(self):
        wf = self.mistral_admin('workflow-validate', params=self.wf_def)

        wf_valid = self.get_field_value(wf, 'Valid')
        wf_error = self.get_field_value(wf, 'Error')

        self.assertEqual('True', wf_valid)
        self.assertEqual('None', wf_error)

    def test_workflow_validate_with_invalid_def(self):
        self.create_file('wf.yaml', 'name: wf\n')

        wf = self.mistral_admin('workflow-validate', params='wf.yaml')

        wf_valid = self.get_field_value(wf, 'Valid')
        wf_error = self.get_field_value(wf, 'Error')

        self.assertEqual('False', wf_valid)
        self.assertNotEqual('None', wf_error)

    def test_workflow_list_with_filter(self):
        self.workflow_create(self.wf_def)
        workflows = self.parser.listing(self.mistral('workflow-list'))

        self.assertTableStruct(
            workflows,
            ['ID', 'Name', 'Tags', 'Input', 'Scope', 'Created at',
             'Updated at']
        )

        # We created 2 workflows, so we should have at least 2
        self.assertGreaterEqual(len(workflows), 2)

        # Now let's provide a filter to the list command.
        workflows = self.parser.listing(
            self.mistral(
                'workflow-list',
                params='--filter name=eq:wf1'
            )
        )

        self.assertTableStruct(
            workflows,
            ['ID', 'Name', 'Tags', 'Input', 'Scope', 'Created at',
             'Updated at']
        )

        self.assertEqual(1, len(workflows))

        self.assertEqual('wf1', workflows[0]['Name'])


class ExecutionCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with executions."""

    @classmethod
    def setUpClass(cls):
        super(ExecutionCLITests, cls).setUpClass()

    def setUp(self):
        super(ExecutionCLITests, self).setUp()

        wfs = self.workflow_create(self.wf_def)
        self.async_wf = self.workflow_create(self.async_wf_def)[0]

        self.direct_wf = wfs[0]
        self.reverse_wf = wfs[1]

        self.create_file('input', '{\n    "farewell": "Bye"\n}\n')
        self.create_file('task_name', '{\n    "task_name": "goodbye"\n}\n')

    def test_execution_by_id_of_workflow_within_namespace(self):
        namespace = 'abc'

        wfs = self.workflow_create(self.lowest_level_wf, namespace=namespace)

        wf_def_name = wfs[0]['Name']
        wf_id = wfs[0]['ID']

        execution = self.execution_create(wf_id)

        self.assertTableStruct(execution, ['Field', 'Value'])

        wf_name = self.get_field_value(execution, 'Workflow name')
        wf_namespace = self.get_field_value(execution, 'Workflow namespace')
        wf_id = self.get_field_value(execution, 'Workflow ID')

        self.assertEqual(wf_def_name, wf_name)
        self.assertEqual(namespace, wf_namespace)
        self.assertIsNotNone(wf_id)

    def test_execution_within_namespace_create_delete(self):
        namespace = 'abc'

        self.workflow_create(self.lowest_level_wf)
        self.workflow_create(self.lowest_level_wf, namespace=namespace)
        self.workflow_create(self.middle_wf, namespace=namespace)
        self.workflow_create(self.top_level_wf)
        wfs = self.workflow_create(self.top_level_wf, namespace=namespace)

        top_wf_name = wfs[0]['Name']
        execution = self.mistral_admin(
            'execution-create',
            params='{0} --namespace {1}'.format(top_wf_name, namespace)
        )
        exec_id = self.get_field_value(execution, 'ID')

        self.assertTableStruct(execution, ['Field', 'Value'])

        wf_name = self.get_field_value(execution, 'Workflow name')
        wf_namespace = self.get_field_value(execution, 'Workflow namespace')
        wf_id = self.get_field_value(execution, 'Workflow ID')
        created_at = self.get_field_value(execution, 'Created at')

        self.assertEqual(top_wf_name, wf_name)
        self.assertEqual(namespace, wf_namespace)
        self.assertIsNotNone(wf_id)
        self.assertIsNotNone(created_at)

        execs = self.mistral_admin('execution-list')

        self.assertIn(exec_id, [ex['ID'] for ex in execs])
        self.assertIn(wf_name, [ex['Workflow name'] for ex in execs])
        self.assertIn(namespace, [ex['Workflow namespace'] for ex in execs])

        params = "{} --force".format(exec_id)
        self.mistral_admin('execution-delete', params=params)

    def test_execution_create_delete(self):

        execution = self.mistral_admin(
            'execution-create',
            params='{0} -d "execution test"'.format(self.direct_wf['Name'])
        )

        exec_id = self.get_field_value(execution, 'ID')

        self.assertTableStruct(execution, ['Field', 'Value'])

        wf_name = self.get_field_value(execution, 'Workflow name')
        wf_id = self.get_field_value(execution, 'Workflow ID')
        created_at = self.get_field_value(execution, 'Created at')
        description = self.get_field_value(execution, 'Description')

        self.assertEqual(self.direct_wf['Name'], wf_name)
        self.assertIsNotNone(wf_id)
        self.assertIsNotNone(created_at)
        self.assertEqual("execution test", description)

        execs = self.mistral_admin('execution-list')

        self.assertIn(exec_id, [ex['ID'] for ex in execs])
        self.assertIn(wf_name, [ex['Workflow name'] for ex in execs])

        params = "{} --force".format(exec_id)
        self.mistral_admin('execution-delete', params=params)

    def test_execution_create_with_input_and_start_task(self):
        execution = self.execution_create(
            "%s input task_name" % self.reverse_wf['Name']
        )

        exec_id = self.get_field_value(execution, 'ID')

        result = self.wait_execution_success(exec_id)

        self.assertTrue(result)

    def test_execution_update(self):
        execution = self.execution_create(self.async_wf['Name'])

        exec_id = self.get_field_value(execution, 'ID')
        status = self.get_field_value(execution, 'State')

        self.assertEqual('RUNNING', status)

        # Update execution state.
        execution = self.mistral_admin(
            'execution-update',
            params='{0} -s PAUSED'.format(exec_id))

        updated_exec_id = self.get_field_value(execution, 'ID')
        status = self.get_field_value(execution, 'State')

        self.assertEqual(exec_id, updated_exec_id)
        self.assertEqual('PAUSED', status)

        # Update execution description.
        execution = self.mistral_admin(
            'execution-update',
            params='{0} -d "execution update test"'.format(exec_id)
        )

        description = self.get_field_value(execution, 'Description')

        self.assertEqual("execution update test", description)

    def test_execution_get(self):
        execution = self.execution_create(self.direct_wf['Name'])

        exec_id = self.get_field_value(execution, 'ID')

        execution = self.mistral_admin(
            'execution-get',
            params='{0}'.format(exec_id)
        )

        gotten_id = self.get_field_value(execution, 'ID')
        wf_name = self.get_field_value(execution, 'Workflow name')
        wf_id = self.get_field_value(execution, 'Workflow ID')

        self.assertIsNotNone(wf_id)
        self.assertEqual(exec_id, gotten_id)
        self.assertEqual(self.direct_wf['Name'], wf_name)

    def test_execution_get_input(self):
        execution = self.execution_create(self.direct_wf['Name'])
        exec_id = self.get_field_value(execution, 'ID')

        ex_input = self.mistral_admin('execution-get-input', params=exec_id)

        self.assertEqual([], ex_input)

    def test_execution_get_output(self):
        execution = self.execution_create(self.direct_wf['Name'])

        exec_id = self.get_field_value(execution, 'ID')

        ex_output = self.mistral_admin('execution-get-output', params=exec_id)

        self.assertEqual([], ex_output)

    def test_executions_list_with_task(self):
        wrapping_wf = self.workflow_create(self.wf_wrapping_wf)
        decoy = self.execution_create(wrapping_wf[-1]['Name'])
        wrapping_wf_ex = self.execution_create(wrapping_wf[-1]['Name'])

        wrapping_wf_ex_id = self.get_field_value(wrapping_wf_ex, 'ID')

        self.assertIsNot(wrapping_wf_ex_id, self.get_field_value(decoy, 'ID'))

        tasks = self.mistral_admin('task-list', params=wrapping_wf_ex_id)

        wrapping_task_id = tasks[-1]['ID']

        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params="--task {}".format(wrapping_task_id)
        )

        self.assertEqual(1, len(wf_execs))

        wf_exec = wf_execs[0]

        self.assertEqual(wrapping_task_id, wf_exec['Task Execution ID'])

    def test_executions_list_with_pagination(self):
        # Create 2 executions
        ex1 = self.execution_create(
            params='{0} -d "a"'.format(self.direct_wf['Name'])
        )
        # We need to sleep at least one sec between the creation of the two
        # executions because the default sort is by created_at descending
        # (newest first)
        # And the tests may fail if the created_at are identical (some sort
        # of race condition)
        time.sleep(1)
        ex2 = self.execution_create(
            params='{0} -d "b"'.format(self.direct_wf['Name'])
        )

        all_wf_ids = [
            self.get_field_value(ex1, 'ID'),
            self.get_field_value(ex2, 'ID'),
        ]

        # List all executions
        wf_execs = self.mistral_cli(True, 'execution-list')

        # We are supposed to have 2
        self.assertEqual(2, len(wf_execs))

        # We are supposed to have the correct IDs
        self.assertEqual(
            set(all_wf_ids),
            set([ex['ID'] for ex in wf_execs])
        )

        # List executions with limit 1
        # NOTE(arnaud) mistral client is returning newest first by default
        # even if it say differently in help message
        # See change: I002edd1b10ab281072cfa7501cfa763073a7781c
        # So here, for the test with marker to work, we need to
        # specificaly ask for ascending execution list using --oldest
        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params="--oldest --limit 1"
        )

        # We are supposed to have one
        self.assertEqual(1, len(wf_execs))

        # List executions starting after the one we received before
        not_expected = wf_execs[0]['ID']
        expected = [ex for ex in all_wf_ids if ex != wf_execs[0]['ID']][0]
        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params="--marker %s" % not_expected
        )

        # Check if we have correct ex ID
        self.assertNotIn(
            not_expected,
            [ex['ID'] for ex in wf_execs]
        )
        self.assertIn(
            expected,
            [ex['ID'] for ex in wf_execs]
        )

        # List sorted by Description
        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params="--sort_keys Description"
        )

        # We are supposed to have both
        self.assertEqual(
            set(all_wf_ids),
            set([ex['ID'] for ex in wf_execs])
        )

        # Now check if they are correctly ordered
        # Ascending by default
        wf_ex1_index = -1
        wf_ex2_index = -1
        for idx, ex in enumerate(wf_execs):
            if ex['ID'] == all_wf_ids[0]:
                wf_ex1_index = idx
            elif ex['ID'] == all_wf_ids[1]:
                wf_ex2_index = idx
        self.assertLess(wf_ex1_index, wf_ex2_index)

        # Check if descending is working also
        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params="--sort_keys Description --sort_dirs=desc"
        )

        self.assertEqual(
            set(all_wf_ids),
            set([ex['ID'] for ex in wf_execs])
        )

        wf_ex1_index = -1
        wf_ex2_index = -1
        for idx, ex in enumerate(wf_execs):
            if ex['ID'] == all_wf_ids[0]:
                wf_ex1_index = idx
            elif ex['ID'] == all_wf_ids[1]:
                wf_ex2_index = idx
        self.assertGreater(wf_ex1_index, wf_ex2_index)

    def test_execution_list_with_filter(self):
        wf_ex1 = self.execution_create(
            params='{0} -d "a"'.format(self.direct_wf['Name'])
        )

        wf_ex1_id = self.get_field_value(wf_ex1, 'ID')

        self.execution_create(
            params='{0} -d "b"'.format(self.direct_wf['Name'])
        )

        # Request a list without filters.
        wf_execs = self.mistral_cli(True, 'execution-list')

        self.assertTableStruct(
            wf_execs,
            ['ID', 'Workflow name', 'Workflow ID', 'State', 'Created at',
             'Updated at']
        )

        self.assertEqual(2, len(wf_execs))

        # Now let's provide a filter.
        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params='--filter description=a'
        )

        self.assertTableStruct(
            wf_execs,
            ['ID', 'Workflow name', 'Workflow ID', 'State', 'Created at',
             'Updated at']
        )

        self.assertEqual(1, len(wf_execs))

        self.assertEqual(wf_ex1_id, wf_execs[0]['ID'])
        self.assertEqual('a', wf_execs[0]['Description'])

    def test_executions_list_with_rootsonly(self):
        wrapping_wf = self.workflow_create(self.wf_wrapping_wf)
        wrapping_wf_ex = self.execution_create(wrapping_wf[-1]['Name'])

        wrapping_wf_ex_id = self.get_field_value(wrapping_wf_ex, 'ID')

        wf_execs = self.mistral_cli(
            True,
            'execution-list',
            params="--rootsonly"
        )

        self.assertEqual(1, len(wf_execs))
        wf_exec = wf_execs[0]
        self.assertEqual(wrapping_wf_ex_id, wf_exec['ID'])


class CronTriggerCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with cron-triggers."""

    @classmethod
    def setUpClass(cls):
        super(CronTriggerCLITests, cls).setUpClass()

    def setUp(self):
        super(CronTriggerCLITests, self).setUp()

        wf = self.workflow_create(self.wf_def)

        self.wf_name = wf[0]['Name']

    def test_cron_trigger_create_delete(self):
        trigger = self.mistral_admin(
            'cron-trigger-create',
            params=('trigger %s {} --pattern "5 * * * *" --count 5'
                    ' --first-time "4242-12-25 13:37" --utc' % self.wf_name)
        )

        self.assertTableStruct(trigger, ['Field', 'Value'])

        tr_name = self.get_field_value(trigger, 'Name')
        wf_name = self.get_field_value(trigger, 'Workflow')
        created_at = self.get_field_value(trigger, 'Created at')
        remain = self.get_field_value(trigger, 'Remaining executions')
        next_time = self.get_field_value(trigger, 'Next execution time')

        self.assertEqual('trigger', tr_name)
        self.assertEqual(self.wf_name, wf_name)
        self.assertIsNotNone(created_at)
        self.assertEqual("4242-12-25 13:37:00", next_time)
        self.assertEqual("5", remain)

        triggers = self.mistral_admin('cron-trigger-list')

        self.assertIn(tr_name, [tr['Name'] for tr in triggers])
        self.assertIn(wf_name, [tr['Workflow'] for tr in triggers])

        self.mistral('cron-trigger-delete', params=tr_name)

        triggers = self.mistral_admin('cron-trigger-list')

        self.assertNotIn(tr_name, [tr['Name'] for tr in triggers])

    def test_two_cron_triggers_for_one_wf(self):
        self.cron_trigger_create('trigger1', self.wf_name, '{}', "5 * * * *")

        self.cron_trigger_create('trigger2', self.wf_name, '{}', "15 * * * *")

        triggers = self.mistral_admin('cron-trigger-list')

        self.assertIn("trigger1", [tr['Name'] for tr in triggers])
        self.assertIn("trigger2", [tr['Name'] for tr in triggers])

    def test_cron_trigger_get(self):
        trigger = self.cron_trigger_create(
            'trigger',
            self.wf_name,
            '{}',
            "5 * * * *"
        )

        self.assertTableStruct(trigger, ['Field', 'Value'])

        fetched_tr = self.mistral_admin(
            'cron-trigger-get',
            params='trigger'
        )

        self.assertTableStruct(trigger, ['Field', 'Value'])

        tr_name = self.get_field_value(fetched_tr, 'Name')
        wf_name = self.get_field_value(fetched_tr, 'Workflow')
        created_at = self.get_field_value(fetched_tr, 'Created at')

        self.assertEqual('trigger', tr_name)
        self.assertEqual(self.wf_name, wf_name)
        self.assertIsNotNone(created_at)


class EventTriggerCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with event-triggers."""

    @classmethod
    def setUpClass(cls):
        super(EventTriggerCLITests, cls).setUpClass()

    def setUp(self):
        super(EventTriggerCLITests, self).setUp()

        wf = self.workflow_create(self.wf_def)

        self.wf_id = wf[0]['ID']

    def test_event_trigger_create_delete(self):
        trigger = self.mistral_admin(
            'event-trigger-create',
            params=('trigger %s dummy_exchange dummy_topic event.dummy {}' %
                    self.wf_id))

        self.assertTableStruct(trigger, ['Field', 'Value'])

        tr_id = self.get_field_value(trigger, 'ID')
        tr_name = self.get_field_value(trigger, 'Name')
        wf_id = self.get_field_value(trigger, 'Workflow ID')
        created_at = self.get_field_value(trigger, 'Created at')

        self.assertEqual('trigger', tr_name)
        self.assertEqual(self.wf_id, wf_id)
        self.assertIsNotNone(created_at)

        triggers = self.mistral_admin('event-trigger-list')

        self.assertIn(tr_name, [tr['Name'] for tr in triggers])
        self.assertIn(wf_id, [tr['Workflow ID'] for tr in triggers])

        self.mistral('event-trigger-delete', params=tr_id)

        triggers = self.mistral_admin('event-trigger-list')

        self.assertNotIn(tr_name, [tr['Name'] for tr in triggers])

    def test_two_event_triggers_for_one_wf(self):
        self.event_trigger_create('trigger1',
                                  self.wf_id,
                                  'dummy_exchange',
                                  'dummy_topic',
                                  'event.dummy',
                                  '{}')

        self.event_trigger_create('trigger2',
                                  self.wf_id,
                                  'dummy_exchange',
                                  'dummy_topic',
                                  'dummy.event',
                                  '{}')

        triggers = self.mistral_admin('event-trigger-list')

        self.assertIn('trigger1', [tr['Name'] for tr in triggers])
        self.assertIn('trigger2', [tr['Name'] for tr in triggers])

    def test_event_trigger_get(self):
        trigger = self.event_trigger_create('trigger',
                                            self.wf_id,
                                            'dummy_exchange',
                                            'dummy_topic',
                                            'event.dummy.other',
                                            '{}')

        self.assertTableStruct(trigger, ['Field', 'Value'])

        ev_tr_id = self.get_field_value(trigger, 'ID')
        fetched_tr = self.mistral_admin('event-trigger-get', params=ev_tr_id)

        self.assertTableStruct(trigger, ['Field', 'Value'])

        tr_name = self.get_field_value(fetched_tr, 'Name')
        wf_id = self.get_field_value(fetched_tr, 'Workflow ID')
        created_at = self.get_field_value(fetched_tr, 'Created at')

        self.assertEqual('trigger', tr_name)
        self.assertEqual(self.wf_id, wf_id)
        self.assertIsNotNone(created_at)


class TaskCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with tasks."""

    def setUp(self):
        super(TaskCLITests, self).setUp()

        wfs = self.workflow_create(self.wf_def)

        self.direct_wf = wfs[0]
        self.reverse_wf = wfs[1]

        self.create_file('input', '{\n    "farewell": "Bye"\n}\n')
        self.create_file('task_name', '{\n    "task_name": "goodbye"\n}\n')

    def test_task_get(self):
        wf_ex = self.execution_create(self.direct_wf['Name'])

        wf_ex_id = self.get_field_value(wf_ex, 'ID')

        tasks = self.mistral_admin('task-list', params=wf_ex_id)

        created_task_id = tasks[-1]['ID']

        fetched_task = self.mistral_admin('task-get', params=created_task_id)
        fetched_task_id = self.get_field_value(fetched_task, 'ID')
        fetched_task_wf_namespace = self.get_field_value(
            fetched_task,
            'Workflow namespace'
        )
        task_execution_id = self.get_field_value(fetched_task,
                                                 'Workflow Execution ID')

        self.assertEqual(created_task_id, fetched_task_id)
        self.assertEqual('', fetched_task_wf_namespace)
        self.assertEqual(wf_ex_id, task_execution_id)

    def test_task_get_list_within_namespace(self):
        namespace = 'aaa'
        self.workflow_create(self.wf_def, namespace=namespace)
        wf_ex = self.execution_create(
            self.direct_wf['Name'] + ' --namespace ' + namespace
        )

        wf_ex_id = self.get_field_value(wf_ex, 'ID')

        tasks = self.mistral_admin('task-list', params=wf_ex_id)

        created_task_id = tasks[-1]['ID']
        created_wf_namespace = tasks[-1]['Workflow namespace']

        fetched_task = self.mistral_admin('task-get', params=created_task_id)
        fetched_task_id = self.get_field_value(fetched_task, 'ID')
        fetched_task_wf_namespace = self.get_field_value(
            fetched_task,
            'Workflow namespace'
        )
        task_execution_id = self.get_field_value(fetched_task,
                                                 'Workflow Execution ID')

        self.assertEqual(created_task_id, fetched_task_id)
        self.assertEqual(namespace, created_wf_namespace)
        self.assertEqual(created_wf_namespace, fetched_task_wf_namespace)
        self.assertEqual(wf_ex_id, task_execution_id)

    def test_task_list_with_filter(self):
        wf_exec = self.execution_create(
            "%s input task_name" % self.reverse_wf['Name']
        )

        exec_id = self.get_field_value(wf_exec, 'ID')

        self.assertTrue(self.wait_execution_success(exec_id))

        # Request task executions without filters.
        tasks = self.parser.listing(self.mistral('task-list'))

        self.assertTableStruct(
            tasks,
            ['ID', 'Name', 'Workflow name', 'Workflow Execution ID', 'State']
        )

        self.assertEqual(2, len(tasks))

        # Now let's provide a filter.
        tasks = self.parser.listing(
            self.mistral(
                'task-list',
                params='--filter name=goodbye'
            )
        )

        self.assertTableStruct(
            tasks,
            ['ID', 'Name', 'Workflow name', 'Workflow Execution ID', 'State']
        )

        self.assertEqual(1, len(tasks))

        self.assertEqual('goodbye', tasks[0]['Name'])

    def test_task_list_with_limit(self):
        wf_exec = self.execution_create(
            "%s input task_name" % self.reverse_wf['Name']
        )

        exec_id = self.get_field_value(wf_exec, 'ID')

        self.assertTrue(self.wait_execution_success(exec_id))

        tasks = self.parser.listing(self.mistral('task-list'))

        tasks = self.parser.listing(
            self.mistral(
                'task-list',
                params='--limit 1'
            )
        )

        self.assertEqual(1, len(tasks))


class ActionCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with actions."""

    @classmethod
    def setUpClass(cls):
        super(ActionCLITests, cls).setUpClass()

    def test_action_create_delete(self):
        init_acts = self.mistral_admin('action-create', params=self.act_def)

        self.assertTableStruct(
            init_acts,
            [
                'Name', 'Is system', 'Input', 'Description', 'Tags',
                'Created at', 'Updated at'
            ]
        )

        self.assertIn('greeting', [action['Name'] for action in init_acts])
        self.assertIn('farewell', [action['Name'] for action in init_acts])

        action_1 = self.get_item_info(
            get_from=init_acts,
            get_by='Name',
            value='greeting'
        )
        action_2 = self.get_item_info(
            get_from=init_acts,
            get_by='Name',
            value='farewell'
        )

        self.assertEqual('<none>', action_1['Tags'])
        self.assertEqual('<none>', action_2['Tags'])

        self.assertEqual('False', action_1['Is system'])
        self.assertEqual('False', action_2['Is system'])

        self.assertEqual('name', action_1['Input'])
        self.assertEqual('None', action_2['Input'])

        acts = self.mistral_admin('action-list')

        self.assertIn(action_1['Name'], [action['Name'] for action in acts])
        self.assertIn(action_2['Name'], [action['Name'] for action in acts])

        self.mistral_admin(
            'action-delete',
            params='{0}'.format(action_1['Name'])
        )
        self.mistral_admin(
            'action-delete',
            params='{0}'.format(action_2['Name'])
        )

        acts = self.mistral_admin('action-list')

        self.assertNotIn(action_1['Name'], [action['Name'] for action in acts])
        self.assertNotIn(action_2['Name'], [action['Name'] for action in acts])

    def test_action_update(self):
        actions = self.action_create(self.act_def)

        created_action = self.get_item_info(
            get_from=actions,
            get_by='Name',
            value='greeting'
        )

        actions = self.mistral_admin('action-update', params=self.act_def)

        updated_action = self.get_item_info(
            get_from=actions,
            get_by='Name',
            value='greeting'
        )

        self.assertEqual(
            created_action['Created at'].split(".")[0],
            updated_action['Created at']
        )
        self.assertEqual(created_action['Name'], updated_action['Name'])
        self.assertEqual(
            created_action['Updated at'],
            updated_action['Updated at']
        )

        actions = self.mistral_admin('action-update', params=self.act_tag_def)

        updated_action = self.get_item_info(
            get_from=actions,
            get_by='Name',
            value='greeting'
        )

        self.assertEqual('tag, tag1', updated_action['Tags'])
        self.assertEqual(
            created_action['Created at'].split(".")[0],
            updated_action['Created at']
        )
        self.assertEqual(created_action['Name'], updated_action['Name'])
        self.assertNotEqual(
            created_action['Updated at'],
            updated_action['Updated at']
        )

    def test_action_update_with_id(self):
        acts = self.action_create(self.act_def)

        created_action = self.get_item_info(
            get_from=acts,
            get_by='Name',
            value='greeting'
        )

        action_id = created_action['ID']

        params = '{0} --id {1}'.format(self.act_tag_def, action_id)
        acts = self.mistral_admin('action-update', params=params)

        updated_action = self.get_item_info(
            get_from=acts,
            get_by='ID',
            value=action_id
        )

        self.assertEqual(
            created_action['Created at'].split(".")[0],
            updated_action['Created at']
        )
        self.assertEqual(created_action['Name'], updated_action['Name'])
        self.assertNotEqual(
            created_action['Updated at'],
            updated_action['Updated at']
        )

    def test_action_update_truncate_input(self):
        input_value = "very_long_input_parameter_name_that_should_be_truncated"

        act_def = """
        version: "2.0"
        action1:
          input:
            - {0}
          base: std.noop
        """.format(input_value)

        self.create_file('action.yaml', act_def)
        self.action_create('action.yaml')

        updated_act = self.mistral_admin('action-update', params='action.yaml')

        updated_act_info = self.get_item_info(
            get_from=updated_act,
            get_by='Name',
            value='action1'
        )

        self.assertEqual(updated_act_info['Input'][:-3], input_value[:25])

    def test_action_get_definition(self):
        self.action_create(self.act_def)

        definition = self.mistral_admin(
            'action-get-definition',
            params='greeting'
        )

        self.assertNotIn('404 Not Found', definition)

    def test_action_get_definition_with_namespace(self):
        self.action_create(self.act_def)

        definition = self.mistral_admin(
            'action-get-definition',
            params='greeting --namespace test_namespace'
        )

        self.assertNotIn('404 Not Found', definition)

    def test_action_get_with_name(self):
        created = self.action_create(self.act_def)

        action_name = created[0]['Name']

        fetched = self.mistral_admin('action-get', params=action_name)

        fetched_action_name = self.get_field_value(fetched, 'Name')

        self.assertEqual(action_name, fetched_action_name)

    def test_action_list_with_filter(self):
        actions = self.parser.listing(self.mistral('action-list'))

        self.assertTableStruct(
            actions,
            ['Name', 'Is system', 'Input', 'Description',
             'Tags', 'Created at', 'Updated at']
        )

        # NOTE(rakhmerov): This length isn't really a number of actions.
        # The problem is that one entity in a table may be on more than
        # one lines depending on their data. For example, for the
        # workflows that we use in our tests it works fine and parsing
        # algorithm is able to parse entities correctly even if they are
        # on multiple lines, but for actions it doesn't. So the only thing
        # we can do is only check if unfiltered table is bigger than
        # filtered.
        # We need to think how to improve it.
        unfiltered_len = len(actions)

        self.assertGreater(unfiltered_len, 0)

        # Now let's provide a filter to the list command.
        actions = self.parser.listing(
            self.mistral(
                'action-list',
                params='--filter name=in:std.echo,std.noop'
            )
        )

        self.assertTableStruct(
            actions,
            ['Name', 'Is system', 'Input', 'Description',
             'Tags', 'Created at', 'Updated at']
        )

        self.assertGreater(unfiltered_len, len(actions))

        action_names = [a['Name'] for a in actions]

        self.assertIn('std.echo', action_names)
        self.assertIn('std.noop', action_names)
        self.assertNotIn('std.ssh', action_names)


class EnvironmentCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with environments."""

    def setUp(self):
        super(EnvironmentCLITests, self).setUp()

        self.create_file(
            'env.yaml',
            'name: env\n'
            'description: Test env\n'
            'variables:\n'
            '  var: "value"'
        )

    def test_environment_create(self):
        env = self.mistral_admin('environment-create', params='env.yaml')

        env_name = self.get_field_value(env, 'Name')
        env_desc = self.get_field_value(env, 'Description')

        self.assertTableStruct(env, ['Field', 'Value'])

        envs = self.mistral_admin('environment-list')

        self.assertIn(env_name, [en['Name'] for en in envs])
        self.assertIn(env_desc, [en['Description'] for en in envs])

        self.mistral_admin('environment-delete', params=env_name)

        envs = self.mistral_admin('environment-list')

        self.assertNotIn(env_name, [en['Name'] for en in envs])

    def test_environment_create_without_description(self):
        self.create_file(
            'env_without_des.yaml',
            'name: env\n'
            'variables:\n'
            '  var: "value"'
        )

        env = self.mistral_admin(
            'environment-create',
            params='env_without_des.yaml'
        )

        env_name = self.get_field_value(env, 'Name')
        env_desc = self.get_field_value(env, 'Description')

        self.assertTableStruct(env, ['Field', 'Value'])

        envs = self.mistral_admin('environment-list')

        self.assertIn(env_name, [en['Name'] for en in envs])
        self.assertIn(env_desc, 'None')

        self.mistral_admin('environment-delete', params='env')

        envs = self.mistral_admin('environment-list')

        self.assertNotIn(env_name, [en['Name'] for en in envs])

    def test_environment_update(self):
        env = self.environment_create('env.yaml')

        env_name = self.get_field_value(env, 'Name')
        env_desc = self.get_field_value(env, 'Description')
        env_created_at = self.get_field_value(env, 'Created at')
        env_updated_at = self.get_field_value(env, 'Updated at')

        self.assertIsNotNone(env_created_at)
        self.assertEqual('None', env_updated_at)

        self.create_file(
            'env_upd.yaml',
            'name: env\n'
            'description: Updated env\n'
            'variables:\n'
            '  var: "value"'
        )

        env = self.mistral_admin('environment-update', params='env_upd.yaml')

        self.assertTableStruct(env, ['Field', 'Value'])

        updated_env_name = self.get_field_value(env, 'Name')
        updated_env_desc = self.get_field_value(env, 'Description')
        updated_env_created_at = self.get_field_value(env, 'Created at')
        updated_env_updated_at = self.get_field_value(env, 'Updated at')

        self.assertEqual(env_name, updated_env_name)
        self.assertNotEqual(env_desc, updated_env_desc)
        self.assertEqual('Updated env', updated_env_desc)
        self.assertEqual(env_created_at.split('.')[0], updated_env_created_at)
        self.assertIsNotNone(updated_env_updated_at)

    def test_environment_get(self):
        env = self.environment_create('env.yaml')

        env_name = self.get_field_value(env, 'Name')
        env_desc = self.get_field_value(env, 'Description')

        env = self.mistral_admin('environment-get', params=env_name)

        fetched_env_name = self.get_field_value(env, 'Name')
        fetched_env_desc = self.get_field_value(env, 'Description')

        self.assertTableStruct(env, ['Field', 'Value'])
        self.assertEqual(env_name, fetched_env_name)
        self.assertEqual(env_desc, fetched_env_desc)

    def test_environment_get_export(self):
        env = self.environment_create('env.yaml')

        env_name = self.get_field_value(env, 'Name')
        env_desc = self.get_field_value(env, 'Description')

        env = self.mistral_admin('environment-get',
                                 params='--export {0}'.format(env_name))

        fetched_env_name = self.get_field_value(env, 'name')
        fetched_env_desc = self.get_field_value(env, 'description')

        self.assertTableStruct(env, ['Field', 'Value'])
        self.assertEqual(env_name, fetched_env_name)
        self.assertEqual(env_desc, fetched_env_desc)


class ActionExecutionCLITests(base_v2.MistralClientTestBase):
    """Test suite checks commands to work with action executions."""

    def setUp(self):
        super(ActionExecutionCLITests, self).setUp()

        wfs = self.workflow_create(self.wf_def)

        self.direct_wf = wfs[0]

        direct_wf_exec = self.execution_create(self.direct_wf['Name'])

        self.direct_ex_id = self.get_field_value(direct_wf_exec, 'ID')

    def test_act_execution_get(self):
        self.wait_execution_success(self.direct_ex_id)

        task = self.mistral_admin('task-list', params=self.direct_ex_id)[0]

        act_ex_from_list = self.mistral_admin(
            'action-execution-list',
            params=task['ID']
        )[0]

        act_ex = self.mistral_admin(
            'action-execution-get',
            params=act_ex_from_list['ID']
        )

        wf_name = self.get_field_value(act_ex, 'Workflow name')
        state = self.get_field_value(act_ex, 'State')

        self.assertEqual(
            act_ex_from_list['ID'],
            self.get_field_value(act_ex, 'ID')
        )
        self.assertEqual(self.direct_wf['Name'], wf_name)
        self.assertEqual('SUCCESS', state)

    def test_act_execution_list_with_limit(self):
        self.wait_execution_success(self.direct_ex_id)

        act_execs = self.mistral_admin('action-execution-list')

        # The workflow execution started in setUp()
        # generates 2 action executions.
        self.assertGreater(len(act_execs), 1)

        act_execs = self.mistral_admin(
            'action-execution-list',
            params="--limit 1"
        )

        self.assertEqual(len(act_execs), 1)

        act_ex = act_execs[0]

        self.assertEqual(self.direct_wf['Name'], act_ex['Workflow name'])
        self.assertEqual('SUCCESS', act_ex['State'])

    def test_act_execution_get_list_within_namespace(self):
        namespace = 'bbb'
        self.workflow_create(self.wf_def, namespace=namespace)
        wf_ex = self.execution_create(
            self.direct_wf['Name'] + ' --namespace ' + namespace
        )
        exec_id = self.get_field_value(wf_ex, 'ID')
        self.wait_execution_success(exec_id)
        task = self.mistral_admin('task-list', params=exec_id)[0]

        act_ex_from_list = self.mistral_admin(
            'action-execution-list',
            params=task['ID']
        )[0]

        act_ex = self.mistral_admin(
            'action-execution-get',
            params=act_ex_from_list['ID']
        )

        wf_name = self.get_field_value(act_ex, 'Workflow name')
        wf_namespace = self.get_field_value(act_ex, 'Workflow namespace')
        status = self.get_field_value(act_ex, 'State')

        self.assertEqual(
            act_ex_from_list['ID'],
            self.get_field_value(act_ex, 'ID')
        )
        self.assertEqual(self.direct_wf['Name'], wf_name)
        self.assertEqual('SUCCESS', status)
        self.assertEqual(namespace, wf_namespace)
        self.assertEqual(namespace, act_ex_from_list['Workflow namespace'])

    def test_act_execution_create_delete(self):
        action_ex = self.mistral_admin(
            'run-action',
            params="std.echo '{0}' --save-result".format(
                '{"output": "Hello!"}')
        )

        action_ex_id = self.get_field_value(action_ex, 'ID')

        self.assertTableStruct(action_ex, ['Field', 'Value'])

        name = self.get_field_value(action_ex, 'Name')
        wf_name = self.get_field_value(action_ex, 'Workflow name')
        task_name = self.get_field_value(action_ex, 'Task name')

        self.assertEqual('std.echo', name)
        self.assertEqual('None', wf_name)
        self.assertEqual('None', task_name)

        action_exs = self.mistral_admin('action-execution-list')

        self.assertIn(action_ex_id, [ex['ID'] for ex in action_exs])

        self.mistral_admin('action-execution-delete', params=action_ex_id)

        action_exs = self.mistral_admin('action-execution-list')

        self.assertNotIn(action_ex_id, [ex['ID'] for ex in action_exs])


class NegativeCLITests(base_v2.MistralClientTestBase):
    """This class contains negative tests."""

    def test_wb_list_extra_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-list',
            params='param'
        )

    def test_wb_get_unexist_wb(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-get',
            params='wb'
        )

    def test_wb_get_without_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-get'
        )

    def test_wb_create_same_name(self):
        self.workbook_create(self.wb_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.workbook_create,
            self.wb_def
        )

    def test_wb_create_with_wrong_path_to_definition(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook_create',
            'wb'
        )

    def test_wb_delete_unexist_wb(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-delete',
            params='wb'
        )

    def test_wb_update_wrong_path_to_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-update',
            params='wb'
        )

    def test_wb_update_nonexistant_wb(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-update',
            params=self.wb_with_tags_def
        )

    def test_wb_create_empty_def(self):
        self.create_file('empty')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-create',
            params='empty'
        )

    def test_wb_update_empty_def(self):
        self.create_file('empty')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-update',
            params='empty'
        )

    def test_wb_get_definition_unexist_wb(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-get-definition',
            params='wb'
        )

    def test_wb_create_invalid_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-create',
            params=self.wf_def
        )

    def test_wb_update_invalid_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-update',
            params=self.wf_def
        )

    def test_wb_update_without_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workbook-update'
        )

    def test_wf_list_extra_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-list',
            params='param'
        )

    def test_wf_get_unexist_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-get',
            params='wf'
        )

    def test_wf_get_without_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-get'
        )

    def test_wf_create_without_definition(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-create',
            params=''
        )

    def test_wf_create_with_wrong_path_to_definition(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-create',
            params='wf'
        )

    def test_wf_delete_unexist_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-delete',
            params='wf'
        )

    def test_wf_update_unexist_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-update',
            params='wf'
        )

    def test_wf_get_definition_unexist_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-get-definition',
            params='wf'
        )

    def test_wf_get_definition_missed_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-get-definition'
        )

    def test_wf_create_invalid_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-create',
            params=self.wb_def
        )

    def test_wf_update_invalid_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-update',
            params=self.wb_def
        )

    def test_wf_create_empty_def(self):
        self.create_file('empty')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-create',
            params='empty'
        )

    def test_wf_update_empty_def(self):
        self.create_file('empty')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'workflow-update',
            params='empty'
        )

    def test_ex_list_extra_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-list',
            params='param'
        )

    def test_ex_create_unexist_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-create',
            params='wf'
        )

    def test_ex_create_unexist_task(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-create',
            params='%s param {}' % wf[0]['Name']
        )

    def test_ex_create_with_invalid_input(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-create',
            params="%s input" % wf[0]['Name']
        )

    def test_ex_get_nonexist_execution(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-get',
            params='%s id' % wf[0]['Name']
        )

    def test_ex_create_without_wf_name(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-create'
        )

    def test_ex_create_reverse_wf_without_start_task(self):
        wf = self.workflow_create(self.wf_def)

        self.create_file('input', '{\n    "farewell": "Bye"\n}\n')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-create ',
            params=wf[1]['Name']
        )

    def test_ex_create_missed_input(self):
        self.create_file('empty')

        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-create empty',
            params=wf[1]['Name']
        )

    def test_ex_update_both_state_and_description(self):
        wf = self.workflow_create(self.wf_def)
        execution = self.execution_create(params=wf[0]['Name'])

        exec_id = self.get_field_value(execution, 'ID')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-update',
            params='%s -s ERROR -d update' % exec_id
        )

    def test_ex_delete_nonexistent_execution(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution-delete',
            params='1a2b3c'
        )

    def test_tr_create_without_pattern(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr %s {}' % wf[0]['Name']
        )

    def test_tr_create_invalid_pattern(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr %s {} --pattern "q"' % wf[0]['Name']
        )

    def test_tr_create_invalid_pattern_value_out_of_range(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr %s {} --pattern "80 * * * *"' % wf[0]['Name']
        )

    def test_tr_create_nonexistent_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr wb.wf1 {} --pattern "* * * * *"'
        )

    def test_tr_delete_nonexistant_tr(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-delete',
            params='tr'
        )

    def test_tr_get_nonexistant_tr(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-get',
            params='tr'
        )

    def test_tr_create_invalid_count(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr wb.wf1 {} --pattern "* * * * *" --count q'
        )

    def test_tr_create_negative_count(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr wb.wf1 {} --pattern "* * * * *" --count -1')

    def test_tr_create_invalid_first_date(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr wb.wf1 {} --pattern "* * * * *" --first-date "q"'
        )

    def test_tr_create_count_only(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr wb.wf1 {} --count 42'
        )

    def test_tr_create_date_and_count_without_pattern(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'cron-trigger-create',
            params='tr wb.wf1 {} --count 42 --first-time "4242-12-25 13:37"'
        )

    def test_event_tr_create_missing_argument(self):
        wf = self.workflow_create(self.wf_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'event-trigger-create',
            params='tr %s exchange topic' % wf[0]['ID']
        )

    def test_event_tr_create_nonexistent_wf(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'event-trigger-create',
            params='456 4307362e-4a4a-4021-aa58-0fab23c9c751 '
                   'exchange topic event {} '
        )

    def test_event_tr_delete_nonexistent_tr(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'event-trigger-delete',
            params='789'
        )

    def test_event_tr_get_nonexistent_tr(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'event-trigger-get',
            params='789'
        )

    def test_action_get_nonexistent(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral,
            'action-get',
            params='nonexist'
        )

    def test_action_double_creation(self):
        self.action_create(self.act_def)

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-create',
            params='{0}'.format(self.act_def)
        )

    def test_action_create_without_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-create',
            params=''
        )

    def test_action_create_invalid_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-create',
            params='{0}'.format(self.wb_def)
        )

    def test_action_delete_nonexistent_act(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-delete',
            params='nonexist'
        )

    def test_action_delete_standard_action(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-delete',
            params='heat.events_get'
        )

    def test_action_get_definition_nonexistent_action(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-get-definition',
            params='nonexist'
        )

    def test_task_get_nonexistent_task(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'task-get',
            params='nonexist'
        )

    def test_env_get_without_param(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-get'
        )

    def test_env_get_nonexistent(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-get',
            params='nonexist'
        )

    def test_env_create_same_name(self):
        self.create_file(
            'env.yaml',
            'name: env\n'
            'description: Test env\n'
            'variables:\n'
            '  var: "value"'
        )

        self.environment_create('env.yaml')

        self.assertRaises(
            exceptions.CommandFailed,
            self.environment_create,
            'env.yaml'
        )

    def test_env_create_empty(self):
        self.create_file('env.yaml')

        self.assertRaises(
            exceptions.CommandFailed,
            self.environment_create,
            'env.yaml'
        )

    def test_env_create_with_wrong_path_to_definition(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'execution_create',
            'env'
        )

    def test_env_delete_unexist_env(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-delete',
            params='env'
        )

    def test_env_update_wrong_path_to_def(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-update',
            params='env'
        )

    def test_env_update_empty(self):
        self.create_file('env.yaml')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-update',
            params='env'
        )

    def test_env_update_nonexistant_env(self):
        self.create_file(
            'env.yaml',
            'name: env'
            'variables:\n  var: "value"'
        )

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-update',
            params='env.yaml'
        )

    def test_env_create_without_name(self):
        self.create_file('env.yaml', 'variables:\n  var: "value"')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-create',
            params='env.yaml'
        )

    def test_env_create_without_variables(self):
        self.create_file('env.yaml', 'name: env')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'environment-create',
            params='env.yaml'
        )

    def test_action_execution_get_without_params(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-execution-get'
        )

    def test_action_execution_get_unexistent_obj(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-execution-get',
            params='123456'
        )

    def test_action_execution_update(self):
        wfs = self.workflow_create(self.wf_def)
        direct_wf_exec = self.execution_create(wfs[0]['Name'])

        direct_ex_id = self.get_field_value(direct_wf_exec, 'ID')

        self.assertRaises(
            exceptions.CommandFailed,
            self.mistral_admin,
            'action-execution-update',
            params='%s ERROR' % direct_ex_id
        )

    def test_target_action_execution(self):
        command = (
            '--debug '
            '--os-target-tenant-name={tenantname} '
            '--os-target-username={username} '
            '--os-target-password="{password}" '
            '--os-target-auth-url="{auth_url}" '
            '--target_insecure '
            'run-action std.noop'
        ).format(
            tenantname=self.clients.tenant_name,
            username=self.clients.username,
            password=self.clients.password,
            auth_url=self.clients.uri
        )

        self.mistral_alt_user(cmd=command)



# Copyright 2014 - Mirantis, Inc.
# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import datetime
import textwrap

from osc_lib.command import command
from oslo_utils import timeutils


DEFAULT_LIMIT = 100


class MistralFormatter(metaclass=abc.ABCMeta):
    COLUMNS = []

    @classmethod
    def fields(cls):
        # Column should be a tuple:
        # (<field name>, <field title>, <optional synthetic flag>)
        # If the 3rd value is specified and it's True then
        # the field is synthetic (calculated) and should not be requested
        # from the API client.
        return [c[0] for c in cls.COLUMNS if len(c) == 2 or not c[2]]

    @classmethod
    def headings(cls):
        return [c[1] for c in cls.COLUMNS]

    @classmethod
    def format_list(cls, instance=None):
        return cls.format(instance, lister=True)

    @staticmethod
    def format(instance=None, lister=False):
        pass


class MistralLister(command.Lister, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _get_format_function(self):
        raise NotImplementedError

    def get_parser(self, parsed_args):
        parser = super(MistralLister, self).get_parser(parsed_args)

        parser.add_argument(
            '--marker',
            type=str,
            help='The last execution uuid of the previous page, displays list '
                 'of executions after "marker".',
            default='',
            nargs='?'
        )
        parser.add_argument(
            '--limit',
            type=int,
            help='Maximum number of entries to return in a single result. ',
            nargs='?'
        )
        parser.add_argument(
            '--sort_keys',
            help='Comma-separated list of sort keys to sort results by. '
                 'Default: created_at. '
                 'Example: mistral execution-list --sort_keys=id,description',
            default='created_at',
            nargs='?'
        )
        parser.add_argument(
            '--sort_dirs',
            help='Comma-separated list of sort directions. Default: asc. '
                 'Example: mistral execution-list --sort_keys=id,description '
                 '--sort_dirs=asc,desc',
            default='asc',
            nargs='?'
        )
        parser.add_argument(
            '--filter',
            dest='filters',
            action='append',
            help='Filters. Can be repeated.'
        )

        return parser

    @abc.abstractmethod
    def _get_resources(self, parsed_args):
        """Gets a list of API resources (e.g. using client)."""
        raise NotImplementedError

    def _validate_parsed_args(self, parsed_args):
        # No-op by default.
        pass

    def take_action(self, parsed_args):
        self._validate_parsed_args(parsed_args)

        f = self._get_format_function()

        ret = self._get_resources(parsed_args)

        if not isinstance(ret, list):
            ret = [ret]

        data = [f(r)[1] for r in ret]

        if data:
            return f()[0], data
        else:
            return f()


class MistralExecutionLister(MistralLister, metaclass=abc.ABCMeta):
    def get_parser(self, parsed_args):
        parser = super(MistralExecutionLister, self).get_parser(parsed_args)
        parser.set_defaults(limit=DEFAULT_LIMIT)
        parser.add_argument(
            '--oldest',
            help='Display the executions starting from the oldest entries '
                 'instead of the newest',
            default=False,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        self._validate_parsed_args(parsed_args)

        f = self._get_format_function()

        reverse_results = False
        if (parsed_args.marker == '' and parsed_args.sort_dirs == 'asc' and
            parsed_args.sort_keys == 'created_at' and
                not parsed_args.oldest):
            reverse_results = True
            parsed_args.sort_dirs = 'desc'

        ret = self._get_resources(parsed_args)
        if not isinstance(ret, list):
            ret = [ret]

        if reverse_results:
            ret.reverse()

        data = [f(r)[1] for r in ret]

        if data:
            return f()[0], data
        else:
            return f()


def cut(string, length=25):
    if string and len(string) > length:
        return "%s..." % string[:length]
    else:
        return string


def wrap(string, width=25):
    if string and len(string) > width:
        return textwrap.fill(string, width)
    else:
        return string


def get_filters(parsed_args):
    filters = {}

    if parsed_args.filters:
        for f in parsed_args.filters:
            arr = f.split('=')

            if len(arr) != 2:
                raise ValueError('Invalid filter: %s' % f)

            filters[arr[0]] = arr[1]

    return filters


def get_duration_str(start_dt_str, end_dt_str):
    """Builds a human friendly duration string.

    :param start_dt_str: Start date time as an ISO string.
    :param end_dt_str: End date time as an ISO string. If empty, duration is
        calculated from the current time.
    :return: Duration(delta) string.
    """
    if not start_dt_str:
        return ''

    start_dt = datetime.datetime.strptime(start_dt_str, '%Y-%m-%d %H:%M:%S')

    if end_dt_str:
        end_dt = datetime.datetime.strptime(end_dt_str, '%Y-%m-%d %H:%M:%S')

        return str(end_dt - start_dt)

    delta_from_now = timeutils.utcnow() - start_dt

    # If delta is too small then we won't show any value. It means that
    # the corresponding process (e.g. an execution) just started.
    if delta_from_now < datetime.timedelta(seconds=2):
        return '...'

    # Drop microseconds to decrease verbosity.
    delta = (
        delta_from_now
        - datetime.timedelta(microseconds=delta_from_now.microseconds)
    )

    return "{}...".format(delta)



# Copyright 2016 Catalyst IT Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy

from mistralclient.tests.unit.v2 import base

MEMBER = {
    'id': '123',
    'resource_id': '456',
    'resource_type': 'workflow',
    'project_id': 'dc4ffdee54d74028b19b1b90e77aa84f',
    'member_id': '04f61e967fa14cd49950ffe2319317ad',
    'status': 'pending',
}

WORKFLOW_MEMBERS_URL = '/workflows/%s/members' % (MEMBER['resource_id'])
WORKFLOW_MEMBER_URL = '/workflows/%s/members/%s' % (
    MEMBER['resource_id'], MEMBER['member_id']
)


class TestWorkflowMembers(base.BaseClientV2Test):
    def test_create(self):
        self.requests_mock.post(self.TEST_URL + WORKFLOW_MEMBERS_URL,
                                json=MEMBER,
                                status_code=201)

        mb = self.members.create(
            MEMBER['resource_id'],
            MEMBER['resource_type'],
            MEMBER['member_id']
        )

        self.assertIsNotNone(mb)

        self.assertDictEqual({'member_id': MEMBER['member_id']},
                             self.requests_mock.last_request.json())

    def test_update(self):
        updated_member = copy.copy(MEMBER)
        updated_member['status'] = 'accepted'

        self.requests_mock.put(self.TEST_URL + WORKFLOW_MEMBER_URL,
                               json=updated_member)

        mb = self.members.update(
            MEMBER['resource_id'],
            MEMBER['resource_type'],
            MEMBER['member_id']
        )

        self.assertIsNotNone(mb)

        self.assertDictEqual({"status": "accepted"},
                             self.requests_mock.last_request.json())

    def test_list(self):
        self.requests_mock.get(self.TEST_URL + WORKFLOW_MEMBERS_URL,
                               json={'members': [MEMBER]})

        mbs = self.members.list(MEMBER['resource_id'], MEMBER['resource_type'])

        self.assertEqual(1, len(mbs))

    def test_get(self):
        self.requests_mock.get(self.TEST_URL + WORKFLOW_MEMBER_URL,
                               json=MEMBER)

        mb = self.members.get(
            MEMBER['resource_id'],
            MEMBER['resource_type'],
            MEMBER['member_id']
        )

        self.assertIsNotNone(mb)

    def test_delete(self):
        self.requests_mock.delete(self.TEST_URL + WORKFLOW_MEMBER_URL,
                                  status_code=204)
        self.members.delete(
            MEMBER['resource_id'],
            MEMBER['resource_type'],
            MEMBER['member_id']
        )



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_serialization import jsonutils

from mistralclient.api import base
from mistralclient.api.v2 import executions


class Task(base.Resource):
    resource_name = 'Task'


class TaskManager(base.ResourceManager):
    resource_class = Task

    def list(self, workflow_execution_id=None, marker='', limit=None,
             sort_keys='', sort_dirs='', fields=None, **filters):
        url = '/tasks'

        if workflow_execution_id:
            url = '/executions/%s/tasks' % workflow_execution_id

        url += '%s'

        query_string = self._build_query_params(
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            fields=fields,
            filters=filters
        )

        return self._list(url % query_string, response_key='tasks')

    def get(self, id):
        self._ensure_not_empty(id=id)

        return self._get('/tasks/%s' % id)

    def get_task_sub_executions(self, id, errors_only='', max_depth=-1):
        task_sub_execs_path = '/tasks/%s/executions%s'
        params = '?max_depth=%s&errors_only=%s' % (max_depth, errors_only)

        return self._list(
            task_sub_execs_path % (id, params),
            response_key='executions',
            returned_res_cls=executions.Execution
        )

    def rerun(self, task_ex_id, reset=True, env=None):
        url = '/tasks/%s' % task_ex_id

        body = {
            'id': task_ex_id,
            'state': 'RUNNING',
            'reset': reset
        }

        if env:
            body['env'] = jsonutils.dumps(env)

        return self._update(url, body)



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock


from mistralclient.api.v2 import code_sources
from mistralclient.commands.v2 import code_sources as code_src_cmd
from mistralclient.tests.unit import base


CODE_SRC_CONTENT = """
from mistral_lib import actions

class HelloAction(actions.Action):
    def __init__(self, f_name, l_name):
        super(HelloAction, self).__init__()

        self._f_name = f_name
        self._l_name = l_name

    def run(self, context):
        return 'Hello %s %s!' % (self._f_name, self._l_name)
"""

CODE_SRC_DICT = {
    'id': '1-2-3-4',
    'name': 'hello_module',
    'namespace': '',
    'project_id': '12345',
    'scope': 'private',
    'content': CODE_SRC_CONTENT,
    'created_at': '1',
    'updated_at': '1'
}

CODE_SRC = code_sources.CodeSource(mock, CODE_SRC_DICT)


class TestCLICodeSources(base.BaseCommandTest):
    @mock.patch('argparse.open', create=True)
    def test_create(self, mock_open):
        self.client.code_sources.create.return_value = CODE_SRC

        result = self.call(
            code_src_cmd.Create,
            app_args=['hello_module', 'hello.py']
        )

        self.assertEqual(
            ('1-2-3-4', 'hello_module', '', '12345', 'private', '1', '1'),
            result[1]
        )

    @mock.patch('argparse.open', create=True)
    def test_create_public(self, mock_open):
        code_src_dict = CODE_SRC_DICT.copy()
        code_src_dict['scope'] = 'public'

        code_src = code_sources.CodeSource(mock, code_src_dict)

        self.client.code_sources.create.return_value = code_src

        result = self.call(
            code_src_cmd.Create,
            app_args=['hello_module', '1.txt', '--public']
        )

        self.assertEqual(
            ('1-2-3-4', 'hello_module', '', '12345', 'public', '1', '1'),
            result[1]
        )

        self.assertEqual(
            'public',
            self.client.code_sources.create.call_args[1]['scope']
        )

    @mock.patch('argparse.open', create=True)
    def test_update(self, mock_open):
        self.client.code_sources.update.return_value = CODE_SRC

        result = self.call(
            code_src_cmd.Update,
            app_args=['hello_module', 'hello.py']
        )

        self.assertEqual(
            ('1-2-3-4', 'hello_module', '', '12345', 'private', '1', '1'),
            result[1]
        )

    @mock.patch('argparse.open', create=True)
    def test_update_public(self, mock_open):
        code_src_dict = CODE_SRC_DICT.copy()
        code_src_dict['scope'] = 'public'

        code_src = code_sources.CodeSource(mock, code_src_dict)

        self.client.code_sources.update.return_value = code_src

        result = self.call(
            code_src_cmd.Update,
            app_args=['hello_module', 'hello.py', '--public']
        )

        self.assertEqual(
            ('1-2-3-4', 'hello_module', '', '12345', 'public', '1', '1'),
            result[1]
        )

        self.assertEqual(
            'public',
            self.client.code_sources.update.call_args[1]['scope']
        )

    @mock.patch('argparse.open', create=True)
    def test_update_with_id(self, mock_open):
        self.client.code_sources.update.return_value = CODE_SRC

        result = self.call(
            code_src_cmd.Update,
            app_args=['1-2-3-4', 'hello.py']
        )

        self.assertEqual(
            ('1-2-3-4', 'hello_module', '', '12345', 'private', '1', '1'),
            result[1]
        )

    def test_list(self):
        self.client.code_sources.list.return_value = [CODE_SRC]

        result = self.call(code_src_cmd.List)

        self.assertEqual(
            [('1-2-3-4', 'hello_module', '', '12345', 'private', '1', '1')],
            result[1]
        )

    def test_get(self):
        self.client.code_sources.get.return_value = CODE_SRC

        result = self.call(code_src_cmd.Get, app_args=['hello_module'])

        self.assertEqual(
            ('1-2-3-4', 'hello_module', '', '12345', 'private', '1', '1'),
            result[1]
        )

    def test_delete(self):
        self.call(code_src_cmd.Delete, app_args=['hello_module'])

        self.client.code_sources.delete.assert_called_once_with(
            'hello_module',
            None
        )

    def test_get_content(self):
        self.client.code_sources.get.return_value = CODE_SRC

        self.call(code_src_cmd.GetContent, app_args=['hello_module'])

        self.app.stdout.write.assert_called_with(CODE_SRC_CONTENT)



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslotest import base
from requests_mock.contrib import fixture


class BaseClientTest(base.BaseTestCase):
    _client = None

    def setUp(self):
        super(BaseClientTest, self).setUp()

        self.requests_mock = self.useFixture(fixture.Fixture())


class BaseCommandTest(base.BaseTestCase):
    def setUp(self):
        super(BaseCommandTest, self).setUp()

        self.app = mock.Mock()
        self.client = self.app.client_manager.workflow_engine

    def call(self, command, app_args=(), prog_name=''):
        cmd = command(self.app, app_args)

        parsed_args = cmd.get_parser(prog_name).parse_args(app_args)

        return cmd.take_action(parsed_args)



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse

from cliff import show
from osc_lib.command import command

from mistralclient.commands.v2 import base
from mistralclient import utils


class CodeSourceFormatter(base.MistralFormatter):
    COLUMNS = [
        ('id', 'ID'),
        ('name', 'Name'),
        ('namespace', 'Namespace'),
        ('project_id', 'Project ID'),
        ('scope', 'Scope'),
        ('created_at', 'Created at'),
        ('updated_at', 'Updated at')
    ]

    @staticmethod
    def format(code_src=None, lister=False):
        if code_src:
            data = (
                code_src.id,
                code_src.name,
                code_src.namespace,
                code_src.project_id,
                code_src.scope,
                code_src.created_at
            )

            if hasattr(code_src, 'updated_at'):
                data += (code_src.updated_at,)
            else:
                data += (None,)
        else:
            data = (('',) * len(CodeSourceFormatter.COLUMNS),)

        return CodeSourceFormatter.headings(), data


class List(base.MistralLister):
    """List all workflows."""

    def _get_format_function(self):
        return CodeSourceFormatter.format_list

    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)

        return parser

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.code_sources.list(
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            sort_keys=parsed_args.sort_keys,
            sort_dirs=parsed_args.sort_dirs,
            fields=CodeSourceFormatter.fields(),
            **base.get_filters(parsed_args)
        )


class Get(show.ShowOne):
    """Show specific code source."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument('identifier', help='Code source ID or name.')
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace to get the code source from.",
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        wf = mistral_client.code_sources.get(
            parsed_args.identifier,
            parsed_args.namespace
        )

        return CodeSourceFormatter.format(wf)


class Create(command.ShowOne):
    """Create new code source."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument('name', help='Code source name.')
        parser.add_argument(
            'content',
            type=argparse.FileType('r'),
            help='Code source content file.'
        )
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace to create the code source within.",
        )
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag the code source will be marked as "public".'
        )

        return parser

    def take_action(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine

        code_source = mistral_client.code_sources.create(
            parsed_args.name,
            parsed_args.content.read(),
            namespace=parsed_args.namespace,
            scope=scope
        )

        return CodeSourceFormatter.format(code_source)


class Delete(command.Command):
    """Delete workflow."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'identifier',
            nargs='+',
            help='Code source name or ID (can be repeated multiple times).'
        )

        parser.add_argument(
            '--namespace',
            nargs='?',
            default=None,
            help="Namespace to delete the code source(s) from.",
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        utils.do_action_on_many(
            lambda s:
                mistral_client.code_sources.delete(s, parsed_args.namespace),
            parsed_args.identifier,
            "Request to delete code source '%s' has been accepted.",
            "Unable to delete the specified code source(s)."
        )


class Update(command.ShowOne):
    """Update workflow."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'identifier',
            help='Code source identifier (name or ID).'
        )
        parser.add_argument(
            'content',
            type=argparse.FileType('r'),
            help='Code source content'
        )
        parser.add_argument('--id', help='Workflow ID.')
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace of the workflow.",
        )
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag workflow will be marked as "public".'
        )

        return parser

    def take_action(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine

        code_src = mistral_client.code_sources.update(
            parsed_args.identifier,
            parsed_args.content.read(),
            namespace=parsed_args.namespace,
            scope=scope
        )

        return CodeSourceFormatter.format(code_src)


class GetContent(command.Command):
    """Show workflow definition."""

    def get_parser(self, prog_name):
        parser = super(GetContent, self).get_parser(prog_name)

        parser.add_argument('identifier', help='Code source ID or name.')
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace to get the code source from.",
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        code_src = mistral_client.code_sources.get(
            parsed_args.identifier,
            parsed_args.namespace
        )

        self.app.stdout.write(code_src.content or "\n")



# Copyright 2015 Huawei Technologies Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import io
import os
import sys

from oslotest import base

from mistralclient import shell


class BaseShellTests(base.BaseTestCase):

    def shell(self, argstr):
        orig = (sys.stdout, sys.stderr)
        clean_env = {}
        _old_env, os.environ = os.environ, clean_env.copy()

        try:
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()
            _shell = shell.MistralShell()
            _shell.run(argstr.split())
        except SystemExit:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.assertEqual(0, exc_value.code)
        finally:
            stdout = sys.stdout.getvalue()
            stderr = sys.stderr.getvalue()
            sys.stdout.close()
            sys.stderr.close()
            sys.stdout, sys.stderr = orig
            os.environ = _old_env

        return stdout, stderr



# Copyright 2015 - Huawei Technologies Co., Ltd.
# Copyright 2016 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import tempfile
from unittest import mock

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from oslotest import base
import osprofiler.profiler

from mistralclient.api import client

AUTH_HTTP_URL_v3 = 'http://localhost:35357/v3'
AUTH_HTTP_URL_v2_0 = 'http://localhost:35357/v2.0'
AUTH_HTTPS_URL = AUTH_HTTP_URL_v3.replace('http', 'https')
MISTRAL_HTTP_URL = 'http://localhost:8989/v2'
MISTRAL_HTTPS_URL = MISTRAL_HTTP_URL.replace('http', 'https')
PROFILER_HMAC_KEY = 'SECRET_HMAC_KEY'


class BaseClientTest(base.BaseTestCase):

    @staticmethod
    def setup_keystone_mock(session_mock):
        keystone_client_instance = session_mock.return_value
        keystone_client_instance.auth_token = uuidutils.generate_uuid()
        keystone_client_instance.project_id = uuidutils.generate_uuid()
        keystone_client_instance.user_id = uuidutils.generate_uuid()
        keystone_client_instance.auth_ref = str(jsonutils.dumps({}))
        return keystone_client_instance

    @mock.patch('keystoneauth1.session.Session')
    def test_mistral_url_from_catalog_v2(self, session_mock):
        session = mock.Mock()
        session_mock.side_effect = [session]

        get_endpoint = mock.Mock(return_value='http://mistral_host:8989/v2')
        session.get_endpoint = get_endpoint

        mistralclient = client.client(
            username='mistral',
            project_name='mistral',
            api_key='password',
            auth_url=AUTH_HTTP_URL_v2_0,
            service_type='workflowv2'
        )

        self.assertEqual(
            'http://mistral_host:8989/v2',
            mistralclient.actions.http_client.base_url
        )

    @mock.patch('keystoneauth1.session.Session')
    def test_mistral_url_from_catalog(self, session_mock):
        session = mock.Mock()
        session_mock.side_effect = [session]

        get_endpoint = mock.Mock(return_value='http://mistral_host:8989/v2')
        session.get_endpoint = get_endpoint

        mistralclient = client.client(
            username='mistral',
            project_name='mistral',
            api_key='password',
            user_domain_name='Default',
            project_domain_name='Default',
            auth_url=AUTH_HTTP_URL_v3,
            service_type='workflowv2'
        )

        self.assertEqual(
            'http://mistral_host:8989/v2',
            mistralclient.actions.http_client.base_url
        )

    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('mistralclient.api.httpclient.HTTPClient')
    def test_mistral_url_default(self, http_client_mock, session_mock):
        session = mock.Mock()
        session_mock.side_effect = [session]

        get_endpoint = mock.Mock(side_effect=Exception)
        session.get_endpoint = get_endpoint

        client.client(
            username='mistral',
            project_name='mistral',
            api_key='password',
            user_domain_name='Default',
            project_domain_name='Default',
            auth_url=AUTH_HTTP_URL_v3
        )

        self.assertTrue(http_client_mock.called)
        mistral_url_for_http = http_client_mock.call_args[0][0]
        self.assertEqual(MISTRAL_HTTP_URL, mistral_url_for_http)

    @mock.patch('mistralclient.auth.keystone.KeystoneAuthHandler'
                '._is_service_catalog_v2', return_value=True)
    @mock.patch('keystoneauth1.identity.generic.Password')
    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('mistralclient.api.httpclient.HTTPClient')
    def test_target_parameters_processed(
        self,
        http_client_mock,
        session_mock,
        password_mock,
        catalog_type_mock
    ):

        session = mock.MagicMock()
        target_session = mock.MagicMock()
        session_mock.side_effect = [session, target_session]
        auth = mock.MagicMock()
        target_auth = mock.MagicMock()
        target_auth._plugin.auth_url = AUTH_HTTP_URL_v3

        password_mock.side_effect = [auth, target_auth]

        get_endpoint = mock.Mock(return_value='http://mistral_host:8989/v2')
        session.get_endpoint = get_endpoint

        target_session.get_project_id = mock.Mock(return_value='projectid')
        target_session.get_user_id = mock.Mock(return_value='userid')
        target_session.get_auth_headers = mock.Mock(return_value={
            'X-Auth-Token': 'authtoken'
        })

        mock_access = mock.MagicMock()
        mock_catalog = mock.MagicMock()
        mock_catalog.catalog = {}
        mock_access.service_catalog = mock_catalog
        auth.get_access = mock.Mock(return_value=mock_access)

        client.client(
            auth_url=AUTH_HTTP_URL_v3,
            username='user',
            api_key='password',
            user_domain_name='Default',
            project_domain_name='Default',
            target_username='tmistral',
            target_project_name='tmistralp',
            target_auth_url=AUTH_HTTP_URL_v3,
            target_api_key='tpassword',
            target_user_domain_name='Default',
            target_project_domain_name='Default',
            target_region_name='tregion'
        )

        self.assertTrue(http_client_mock.called)
        mistral_url_for_http = http_client_mock.call_args[0][0]
        kwargs = http_client_mock.call_args[1]
        self.assertEqual('http://mistral_host:8989/v2', mistral_url_for_http)

        expected_values = {
            'target_project_id': 'projectid',
            'target_auth_token': 'authtoken',
            'target_user_id': 'userid',
            'target_auth_url': AUTH_HTTP_URL_v3,
            'target_project_name': 'tmistralp',
            'target_username': 'tmistral',
            'target_region_name': 'tregion',
            'target_service_catalog': "{}"
        }

        for key in expected_values:
            self.assertEqual(expected_values[key], kwargs[key])

    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('mistralclient.api.httpclient.HTTPClient')
    def test_mistral_url_https_insecure(self, http_client_mock, session_mock):
        keystone_client_instance = self.setup_keystone_mock(  # noqa
            session_mock
        )

        expected_args = (
            MISTRAL_HTTPS_URL,
        )

        client.client(
            mistral_url=MISTRAL_HTTPS_URL,
            username='mistral',
            project_name='mistral',
            api_key='password',
            user_domain_name='Default',
            project_domain_name='Default',
            auth_url=AUTH_HTTP_URL_v3,
            cacert=None,
            insecure=True
        )

        self.assertTrue(http_client_mock.called)
        self.assertEqual(http_client_mock.call_args[0], expected_args)
        self.assertEqual(http_client_mock.call_args[1]['insecure'], True)

    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('mistralclient.api.httpclient.HTTPClient')
    def test_mistral_url_https_secure(self, http_client_mock, session_mock):
        fd, cert_path = tempfile.mkstemp(suffix='.pem')

        keystone_client_instance = self.setup_keystone_mock(  # noqa
            session_mock
        )

        expected_args = (
            MISTRAL_HTTPS_URL,
        )

        try:
            client.client(
                mistral_url=MISTRAL_HTTPS_URL,
                username='mistral',
                project_name='mistral',
                api_key='password',
                user_domain_name='Default',
                project_domain_name='Default',
                auth_url=AUTH_HTTP_URL_v3,
                cacert=cert_path,
                insecure=False
            )
        finally:
            os.close(fd)
            os.unlink(cert_path)

        self.assertTrue(http_client_mock.called)
        self.assertEqual(http_client_mock.call_args[0], expected_args)
        self.assertEqual(http_client_mock.call_args[1]['cacert'], cert_path)

    @mock.patch('keystoneauth1.session.Session')
    def test_mistral_url_https_bad_cacert(self, session_mock):
        keystone_client_instance = self.setup_keystone_mock(  # noqa
            session_mock
        )

        self.assertRaises(
            ValueError,
            client.client,
            mistral_url=MISTRAL_HTTPS_URL,
            username='mistral',
            project_name='mistral',
            api_key='password',
            user_domain_name='Default',
            project_domain_name='Default',
            auth_url=AUTH_HTTP_URL_v3,
            cacert='/path/to/foobar',
            insecure=False
        )

    @mock.patch('logging.Logger.warning')
    @mock.patch('keystoneauth1.session.Session')
    def test_mistral_url_https_bad_insecure(self, session_mock,
                                            log_warning_mock):
        fd, path = tempfile.mkstemp(suffix='.pem')

        keystone_client_instance = self.setup_keystone_mock(
            session_mock
        )

        try:
            client.client(
                mistral_url=MISTRAL_HTTPS_URL,
                user_id=keystone_client_instance.user_id,
                project_id=keystone_client_instance.project_id,
                api_key='password',
                user_domain_name='Default',
                project_domain_name='Default',
                auth_url=AUTH_HTTP_URL_v3,
                cacert=path,
                insecure=True
            )
        finally:
            os.close(fd)
            os.unlink(path)

        self.assertTrue(log_warning_mock.called)

    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('mistralclient.api.httpclient.HTTPClient')
    def test_mistral_profile_enabled(self, http_client_mock, session_mock):
        keystone_client_instance = self.setup_keystone_mock(  # noqa
            session_mock
        )

        client.client(
            username='mistral',
            project_name='mistral',
            api_key='password',
            user_domain_name='Default',
            project_domain_name='Default',
            auth_url=AUTH_HTTP_URL_v3,
            profile=PROFILER_HMAC_KEY
        )

        self.assertTrue(http_client_mock.called)

        profiler = osprofiler.profiler.get()

        self.assertEqual(profiler.hmac_key, PROFILER_HMAC_KEY)

    @mock.patch('mistralclient.auth.get_auth_handler')
    def test_mistral_no_auth(self, get_auth_handler_mock):
        # Test that we don't authenticate if auth url wasn't provided

        client.client(
            username='mistral',
            project_name='mistral',
            api_key='password',
            service_type='workflowv2'
        )

        self.assertEqual(0, get_auth_handler_mock.call_count)



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections
import copy
import importlib.resources

from oslo_serialization import jsonutils

from urllib import parse
from urllib import request

from mistralclient.api import base as api_base
from mistralclient.api.v2 import environments
from mistralclient.tests.unit.v2 import base
from mistralclient import utils


ENVIRONMENT = {
    'name': 'env1',
    'description': 'Test Environment #1',
    'scope': 'private',
    'variables': {
        'server': 'localhost'
    }
}

URL_TEMPLATE = '/environments'
URL_TEMPLATE_NAME = '/environments/%s'


class TestEnvironmentsV2(base.BaseClientV2Test):

    def test_create(self):
        data = copy.deepcopy(ENVIRONMENT)

        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE,
                                status_code=201,
                                json=data)
        env = self.environments.create(**data)

        self.assertIsNotNone(env)

        expected_data = copy.deepcopy(data)
        expected_data['variables'] = jsonutils.dumps(
            expected_data['variables'])

        self.assertEqual(expected_data, self.requests_mock.last_request.json())

    def test_create_with_json_file_uri(self):
        # The contents of env_v2.json must be equivalent to ENVIRONMENT
        path = str(importlib.resources.files('mistralclient').joinpath(
            'tests/unit/resources/env_v2.json'
        ))

        # Convert the file path to file URI
        uri = parse.urljoin('file:', request.pathname2url(path))
        data = collections.OrderedDict(
            utils.load_content(
                utils.get_contents_if_file(uri)
            )
        )

        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE,
                                status_code=201,
                                json=data)
        file_input = {'file': uri}
        env = self.environments.create(**file_input)

        self.assertIsNotNone(env)

        expected_data = copy.deepcopy(data)
        expected_data['variables'] = jsonutils.dumps(
            expected_data['variables'])

        self.assertEqual(expected_data, self.requests_mock.last_request.json())

    def test_create_without_name(self):
        data = copy.deepcopy(ENVIRONMENT)
        data.pop('name')

        e = self.assertRaises(api_base.APIException,
                              self.environments.create,
                              **data)

        self.assertEqual(400, e.error_code)

    def test_update(self):
        data = copy.deepcopy(ENVIRONMENT)

        self.requests_mock.put(self.TEST_URL + URL_TEMPLATE, json=data)
        env = self.environments.update(**data)

        self.assertIsNotNone(env)

        expected_data = copy.deepcopy(data)
        expected_data['variables'] = jsonutils.dumps(
            expected_data['variables'])

        self.assertEqual(expected_data, self.requests_mock.last_request.json())

    def test_update_with_yaml_file(self):
        # The contents of env_v2.json must be equivalent to ENVIRONMENT
        path = str(importlib.resources.files('mistralclient').joinpath(
            'tests/unit/resources/env_v2.json'
        ))
        data = collections.OrderedDict(
            utils.load_content(
                utils.get_contents_if_file(path)
            )
        )
        self.requests_mock.put(self.TEST_URL + URL_TEMPLATE, json=data)

        file_input = {'file': path}
        env = self.environments.update(**file_input)

        self.assertIsNotNone(env)

        expected_data = copy.deepcopy(data)
        expected_data['variables'] = jsonutils.dumps(
            expected_data['variables'])

        self.assertEqual(expected_data, self.requests_mock.last_request.json())

    def test_update_without_name(self):
        data = copy.deepcopy(ENVIRONMENT)
        data.pop('name')

        e = self.assertRaises(api_base.APIException,
                              self.environments.update,
                              **data)

        self.assertEqual(400, e.error_code)

    def test_list(self):
        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE,
                               json={'environments': [ENVIRONMENT]})

        environment_list = self.environments.list()

        self.assertEqual(1, len(environment_list))

        env = environment_list[0]

        self.assertDictEqual(
            environments.Environment(self.environments, ENVIRONMENT).to_dict(),
            env.to_dict()
        )

    def test_get(self):
        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE_NAME % 'env',
                               json=ENVIRONMENT)

        env = self.environments.get('env')

        self.assertIsNotNone(env)
        self.assertDictEqual(
            environments.Environment(self.environments, ENVIRONMENT).to_dict(),
            env.to_dict()
        )

    def test_delete(self):
        self.requests_mock.delete(self.TEST_URL + URL_TEMPLATE_NAME % 'env',
                                  status_code=204)

        self.environments.delete('env')



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from mistralclient.api import base


class Execution(base.Resource):
    resource_name = 'Execution'


class ExecutionManager(base.ResourceManager):
    resource_class = Execution

    def create(self, wf_identifier='', namespace='',
               workflow_input=None, description='', source_execution_id=None,
               **params):
        self._ensure_not_empty(
            workflow_identifier=wf_identifier or source_execution_id
        )

        data = {'description': description}

        if uuidutils.is_uuid_like(source_execution_id):
            data.update({'source_execution_id': source_execution_id})

        if wf_identifier:
            if uuidutils.is_uuid_like(wf_identifier):
                data.update({'workflow_id': wf_identifier})
            else:
                data.update({'workflow_name': wf_identifier})

        if namespace:
            data.update({'workflow_namespace': namespace})

        if workflow_input:
            if isinstance(workflow_input, str):
                data.update({'input': workflow_input})
            else:
                data.update({'input': jsonutils.dumps(workflow_input)})

        if params:
            data.update({'params': jsonutils.dumps(params)})

        return self._create('/executions', data)

    def update(self, id, state, description=None, env=None):
        data = {}

        if state:
            data['state'] = state

        if description:
            data['description'] = description

        if env:
            data['params'] = {'env': env}

        return self._update('/executions/%s' % id, data)

    def list(self, task=None, marker='', limit=None, sort_keys='',
             sort_dirs='', fields='', **filters):
        if task:
            filters['task_execution_id'] = task

        # for backwards compatibility
        if 'nulls' in filters and not filters['nulls']:
            filters.pop('nulls')

        query_string = self._build_query_params(
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            fields=fields,
            filters=filters
        )

        return self._list(
            '/executions%s' % query_string,
            response_key='executions',
        )

    def get(self, id):
        self._ensure_not_empty(id=id)

        return self._get('/executions/%s' % id)

    def get_ex_sub_executions(self, id, errors_only='', max_depth=-1):
        ex_sub_execs_path = '/executions/%s/executions%s'
        params = '?max_depth=%s&errors_only=%s' % (max_depth, errors_only)

        return self._list(
            ex_sub_execs_path % (id, params),
            response_key='executions'
        )

    def delete(self, id, force=None):
        self._ensure_not_empty(id=id)

        query_params = {}

        if force:
            query_params['force'] = True

        query_string = self._build_query_params(filters=query_params)

        self._delete('/executions/%s%s' % (id, query_string))

    def get_report(self, id, errors_only=True, max_depth=None,
                   statistics_only=False):
        self._ensure_not_empty(id=id)

        query_params = {}

        if errors_only:
            query_params['errors_only'] = True

        if statistics_only:
            query_params['statistics_only'] = True

        if max_depth is not None:
            query_params['max_depth'] = max_depth

        query_string = self._build_query_params(filters=query_params)

        resp = self.http_client.get(
            '/executions/%s/report%s' % (id, query_string)
        )

        return resp.json()
